#!/bin/bash
#
# Author Krigler Pavol
# Version 0.3.2
#
# Bash script Cisco Template Manager (CTM) for checking Cisco configuration
# compared to defined templates.
#
# Copyright 2010 Pavol Krigler <pavol.krigler@gelogic.net>
#
# This file is part of Cisco Template Manager.
#
#    CTM is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    CTM is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with CTM.  If not, see <http://www.gnu.org/licenses/>.

# CTM configuration file
CONFIGFILE="/etc/ctm/ctm.conf"

#
# Do not modify variables below
#

# Temporary directory where the configuration file is parsed
TMPDIR="$HOME/.ctm-conf-checker-$$/"

GLOBAL_TEMPLATE_FILE="global.template"
GLOBAL_EXCLUDE_TEMPLATE_FILE="global.exclude"

VERSION="0.3.2"

GLOBAL_E=0
GLOBAL_MODE=1
SECTION_MODE=0
SECTION_E=0
SECTION_ACL_E=0
IFACE_E=0
SECTION_NUM=0
IFACE_NUM=0

PARSED_GLOBAL="$TMPDIR/globalcfg"

GLOBAL_REPORT="$TMPDIR/report-global"
SECTION_REPORT_PREFIX="report-section"
IFACE_REPORT_PREFIX="report-iface"
ACL_REPORT_PREFIX="report-acl"

DEBUGFILE="/tmp/ctm-$VERSION-$$-debug.log"

# Constants
AVAILABLE_TEMPLATES=1

CHECK_IFACE_ETHERNET=1
CHECK_IFACE_SERIAL=1
CHECK_IFACE_VLAN=1
CHECK_IFACE_PORT_CHANNEL=1
CHECK_IFACE_LOOPBACK=0
CHECK_IFACE_TUNNEL=0
CHECK_IFACE_VIRT_TEMPLATE=0

COPY_AND_PASTE_MODE=0
COPY_AND_PASTE_PRE_EXEC[0]=""
COPY_AND_PASTE_POST_EXEC[0]=""

#
# Write debug to log file
#
debug () {
	test -n "$DEBUG" && echo "$(date) :: $1" >> $DEBUGFILE
}

# 
# Show help
#
function show_help() {
	echo -e "Usage: ctm-conf-checker.sh DEVICE ROLE CONFIG [parseonly]\n"\
"Compare Cisco configuration file with defined template. Version $VERSION\n\n"\
"Mandatory arguments:\n"\
"  DEVICE               select device template to compare\n"\
"  ROLE                 select role of the device\n"\
"  CONFIG               path to Cisco configuration file\n\n"\
"Optional argument:\n"\
"  parseonly            CTM will only parse configuration file to separated files"

if [ $AVAILABLE_TEMPLATES -eq 1 ]; then
	echo
	echo "List of available templates and roles:"
	cd $TEMPLATE_DIR

        for ENTRY in $(ls -1); do
                printf "%-18s" $ENTRY; ls -1 $ENTRY | tr '\n' '/' | sed -e 's/\/$//'
                echo
        done
fi
}

#
# Parse cisco configuration file to separate files (sections)
# global-configuration
# interface(s), Physical-ethernet, Vlan, Loopback
# Return: DEV_HOSTNAME
#
parse_config_file() 
{
	DEV_HOSTNAME=""
	local TMP_GLOBAL=""
	local TMP_SECTION=""

	debug "parse_config_file()"

	IFS=$'\n'

	while read LINE
	do
		SECTION_SPACE=${LINE:0:1}

		[ -z "$LINE" ] || [ "$LINE" = '!' ] && continue

		if [ ${LINE:0:12} = "access-list " ]; then
			SECTION_NAME=$(echo $LINE | cut -d " " -f 1,2 | tr ' ' '_')
			echo "$LINE" >> $TMPDIR/$SECTION_NAME
		fi

		if [ ! "$SECTION_SPACE" = ' ' ] && [ ! -z "$TMP_SECTION" ]; then
			echo -e "$TMP_SECTION" | sed -e 's/^ *//' | grep -v '^$' >> $TMPDIR/$SECTION_NAME
			TMP_SECTION=""
			SECTION_NAME=""
		fi

		# Detect new Section
		IS_SECTION=$(echo "$LINE" | egrep -c  "^interface |^router |^line |^route-map |^ipv6 router |^class-map |^policy-map |^ip access-list |^ipv6 access-list |^mac access-list |^object-group ")

		# If section detected, fix name 
		test $IS_SECTION -eq 1 && SECTION_NAME=$(echo $LINE | tr ' /' '_' | sed -e 's/_$//') && touch $TMPDIR/$SECTION_NAME && continue

		# Write section content		
		if [ ! -z "$SECTION_NAME" ] && [ "$SECTION_SPACE" = ' ' ]; then
			TMP_SECTION+="${LINE/ /}\n"
			continue
		fi

		# If no section detected	
		TMP_GLOBAL+="${LINE}\n"

	done < $TMPDIR/source-config-file

	unset IFS

rm -f $TMPDIR/$1

# dump TMP_GLOBAL to file
echo -e "$TMP_GLOBAL" > $PARSED_GLOBAL

# parse device hostname
DEV_HOSTNAME=$(echo -e "$TMP_GLOBAL" | grep "^hostname " | cut -d " " -f 2)
}

#
# Remove last character '\r' from cisco configuration files
#
remove_trailing_characters()
{
	debug "remove_trailing_characters()"
	tr -d '\r' > $TMPDIR/source-config-file < $1
}

#
# Check the global configuration settings
#
check_global_settings()
{
	debug "check_global_settings()"

	if [ ! -e "$DEV_TEMPLATE_DIR/$GLOBAL_TEMPLATE_FILE" ]; then
		echo "ERROR: global template configuration file not found in $DEV_TEMPLATE_DIR/$GLOBAL_TEMPLATE_FILE"
		cleanup $4
		exit
	fi

	while read TEMPLATE_LINE
	do
		# Skip comparation when "!" was found
		if [ -z "$TEMPLATE_LINE" ] || [ ${TEMPLATE_LINE:0:1} = '!' ]; then 
			continue
		fi
 
		RESULT=$(egrep "$TEMPLATE_LINE" $PARSED_GLOBAL)

		if [ -z "$RESULT" ]; then
			GLOBAL_E=$((GLOBAL_E+$(echo "$RESULT" | wc -l)))
			echo "MISSING: $TEMPLATE_LINE" >> $GLOBAL_REPORT 

			debug "check_global_settings() :: Missing line -> $TEMPLATE_LINE"
#		else
#			echo $RESULT >> /tmp/AAA
		fi
	done <  $DEV_TEMPLATE_DIR/$GLOBAL_TEMPLATE_FILE
}

#
# Check the "exclude" in global configuration settings
#
check_global_exclude_settings()
{
	debug "check_global_exclude_settings()"

	if [ ! -e $DEV_TEMPLATE_DIR/$GLOBAL_EXCLUDE_TEMPLATE_FILE ]; then
		echo "notice: global exclude template configuration file $GLOBAL_EXCLUDE_TEMPLATE_FILE not found"
		return
	fi

	debug "check_global_exclude_settings() :: $DEV_TEMPLATE_DIR/$GLOBAL_EXCLUDE_TEMPLATE_FILE"

	while read TEMPLATE_LINE
	do
	
		# Skip comparation when "!" or "empty line" was found
		if [ -z "$TEMPLATE_LINE" ] || [ ${TEMPLATE_LINE:0:1} = '!' ]; then 
			continue
		fi
 
		RESULT=$(egrep "$TEMPLATE_LINE" $PARSED_GLOBAL)
		if [ -n "$RESULT" ]; then
			GLOBAL_E=$((GLOBAL_E+$(echo "$RESULT" | wc -l)))
			echo "$RESULT" | awk '{print "EXCLUDE: " $0}' >> $GLOBAL_REPORT

			debug "check_global_exclude_settings() :: Exclude line -> $RESULT"
		fi
	done <  $DEV_TEMPLATE_DIR/$GLOBAL_EXCLUDE_TEMPLATE_FILE
}

#
# Check sections
#
function check_sections() {
	debug "check_sections()"

	# Check if "sections" directory exists
	if [ ! -e "$DEV_TEMPLATE_DIR/sections" ]; then
		echo "Section subdirectory does not exists, checking skipped"
		return
	fi

	# List all sessions
	SECTION_LIST=$(find "$DEV_TEMPLATE_DIR/sections/" -type f -printf "%f\n" -o -type l -printf "%f\n" | grep -v '.exclude' | grep -v 'access-list')

	for SEARCH in $SECTION_LIST; do
		debug "check_sections() :: Section file $SEARCH"

		WILDCARD=$(echo "$SEARCH" | tr '_' '.')

		debug "check_sections() :: find \"$DEV_TEMPLATE_DIR/sections/\" -type f -name \"$WILDCARD\" -printf \"%f\n\""

		RESULT=$(find "$TMPDIR" -type f -iregex "${TMPDIR}${WILDCARD}$" -printf "%f\n" -o -type l -iregex "${TMPDIR}${WILDCARD}$" -printf "%f\n")

		if [ -z "$RESULT" ]; then
			SECTION_E=$(($SECTION_E + 1))
			if [ $COPY_AND_PASTE_MODE -eq 1 ]; then
				cat $DEV_TEMPLATE_DIR/sections/$SEARCH | sed -e 's/^\(.*\)/MISSING: \1/g' > $TMPDIR/${SECTION_REPORT_PREFIX}-$SEARCH
			else
				echo "MISSING: Missing all section commands" > $TMPDIR/${SECTION_REPORT_PREFIX}-$SEARCH
			fi
			continue
		fi

		for SECTION in $RESULT; do
			SECTION_NUM=$((SECTION_NUM+1))
	
			debug "check_sections() :: SECTION=$SECTION"
				
			# Compare template and configuration section 
			while read TEMPLATE_LINE
			do

				# Skip comparation when "!" was found
				if [ -z "$TEMPLATE_LINE" ] || [ ${TEMPLATE_LINE:0:1} = '!' ]; then 
					continue
				fi

				RESULT=$(egrep "$TEMPLATE_LINE" $TMPDIR/$SECTION)

				if [ -z "$RESULT" ]; then
					SECTION_E=$((SECTION_E+$(echo "$RESULT" | wc -l)))
					echo "MISSING: $TEMPLATE_LINE" >> $TMPDIR/${SECTION_REPORT_PREFIX}-$SECTION
	
					debug "check_sections() :: SECTION=$SECTION :: Missing line -> $TEMPLATE_LINE"
				fi
			done <  $DEV_TEMPLATE_DIR/sections/$SEARCH
		done

	done
}

#
# Check "excluded" commands in sections 
#
function check_sections_exclude() {
	debug "check_sections_exclude()"

	# Check if "sections" directory exists
	if [ ! -e "$DEV_TEMPLATE_DIR/sections" ]; then
#		echo "Section subdirectory does not exists, checking skipped"
		return
	fi

	# List all sessions
	EXC_SECTION_LIST=$(find "$DEV_TEMPLATE_DIR/sections/" -name "*.exclude" -type f -printf "%f\n" | sed -e 's/.exclude$//')

	for SECTION in $EXC_SECTION_LIST; do
		debug "check_sections_exclude() :: Section file $SECTION.exclude"

		if [ ! -e $TMPDIR/$SECTION ]; then
			echo "notice: Section $SECTION not found in configration file, skipping ..."
			continue
		fi
			
		# Compare template and configuration section 
		while read TEMPLATE_LINE
		do
			# Skip comparation when "!" was found
			if [ -z "$TEMPLATE_LINE" ] || [ ${TEMPLATE_LINE:0:1} = '!' ]; then 
				continue
			fi

			RESULT=$(egrep "$TEMPLATE_LINE" $TMPDIR/$SECTION)

			if [ -n "$RESULT" ]; then
				SECTION_E=$((SECTION_E+$(echo "$RESULT" | wc -l)))
#				echo "EXCLUDE: $RESULT" >> $TMPDIR/${SECTION_REPORT_PREFIX}-$SECTION
				echo "$RESULT" | awk '{print "EXCLUDE: " $0}' >> $TMPDIR/${SECTION_REPORT_PREFIX}-$SECTION

				debug "check_sections_exclude() :: SECTION=$SECTION :: Exclude line -> $RESULT"
			fi
		done <  $DEV_TEMPLATE_DIR/sections/$SECTION.exclude
	done
}

#
# Detect interface description and interface status (up/shutdown)
# Parameter 1: Path to parsed interface
# Return values: IFACE_DESC   -> interface description
#                IFACE_STATUS -> interface status (up/shutdown)
#		 IFACE_TYPE   -> type detected in [] marks
#
function parse_iface() {
	debug "parse_iface()"

	# Check if descriptions exists (take first description)
	IFACE_DESC=$(grep -i "description" $1 | head -1 | sed -e 's/description //')
	if [ -n "$IFACE_DESC" ]; then

		debug "parse_iface() :: IFACE_DESC=\"$IFACE_DESC\""

		# Detect interface type
		RESULT=$(echo "$IFACE_DESC" | sed -e 's/.*\[\(.*\)\].*/\1/')
		if [ "$RESULT" != "$IFACE_DESC" ]; then
			IFACE_TYPE="$RESULT"
		else
			IFACE_TYPE="N/A"
		fi
	
		debug "parse_iface() :: IFACE_TYPE=$IFACE_TYPE"
	fi

	# Detect if interface is administratively shutdowned
	RESULT=$(egrep -i '^shutdown *$' $1)
	test -z "$RESULT" && IFACE_STATUS="UP" || IFACE_STATUS="DOWN"

	# EXCEPTION for VLAN1
	if [ $(echo "$1" | egrep -c "Vlan1$") -eq 1 ]; then 
		IFACE_STATUS="UP"
	fi
	
	debug "parse_iface() :: IFACE_STATUS=$IFACE_STATUS :: IFACE_DESC=$IFACE_DESC"

	# Check IFACE_TYPE to "N/A" if missing description on interface
	if [ "$IFACE_STATUS" = "UP" ] && [ -z "$IFACE_DESC" ]; then
		IFACE_TYPE="N/A"
	fi
}

#
# Check the service configuration settings on interfaces
# Parameter 1: Absolute path to parsed interface file
# Parameter 2: Template file (absolute path)
#
check_service_settings()
{
	local IFACE_IP_ADDR=""

	debug "check_service_settings()"

	# check if service template and definition file exists
	if [ ! -e "$DEV_TEMPLATE_DIR/services/$2" ]; then
		echo "warning: template file $2 not found, skipping ..."
		return
	fi

	# Check if ".def" file exists (definition)
	if [ ! -e "$DEV_TEMPLATE_DIR/services/$2.def" ]; then
		echo "warning: definition file $2.def not found, skipping ..."
		return
	else
		# Load definitions values (FORMAT and EXAMPLE)
		. $DEV_TEMPLATE_DIR/services/$2.def

		# Chech if values are defined in DEFINITION file
		if [ -z "$FORMAT" ]; then
			echo "ERROR: FORMAT paramater missing in definition file $DEV_TEMPLATE_DIR/services/$2.def, skipping ..."
			return
		fi 
		if [ -z "$EXAMPLE" ]; then
			echo "ERROR: EXAMPLE paramater missing in definition file $DEV_TEMPLATE_DIR/services/$2.def, skipping ..."
			return
		fi 
	fi

	RESULT=$(echo "$IFACE_DESC" | egrep "^${FORMAT}$")
	if [ -z "$RESULT" ]; then
		echo "MISSING: description $IFACE_DESC" >> $TMPDIR/${IFACE_REPORT_PREFIX}-${IFACE}
		echo "OTHER: Incorrect format of description" >> $TMPDIR/${IFACE_REPORT_PREFIX}-${IFACE}
		echo "OTHER: Your conf: \"description $IFACE_DESC\"" >> $TMPDIR/${IFACE_REPORT_PREFIX}-${IFACE}
		echo "OTHER: Should be: \"description $EXAMPLE\"" >> $TMPDIR/${IFACE_REPORT_PREFIX}-${IFACE}
		IFACE_E=$((IFACE_E+1))
	fi

	while read TEMPLATE_LINE
	do
		test -z "$TEMPLATE_LINE" && continue

		# Skip comparation when "!" or empty line was found
		if [ -z "$TEMPLATE_LINE" ] || [ ${TEMPLATE_LINE:0:1} = '!' ]; then 
			continue
		fi

		RESULT=$(egrep "$TEMPLATE_LINE" $TMPDIR/$1)
		if [ -z "$RESULT" ]; then
			IFACE_E=$((IFACE_E+$(echo "$RESULT" | wc -l)))
			
			# Insert description at the first line
			test ! -e $TMPDIR/${IFACE_REPORT_PREFIX}-${IFACE} && echo "MISSING: ! description $IFACE_DESC" >> $TMPDIR/${IFACE_REPORT_PREFIX}-${IFACE}

			echo "MISSING: $TEMPLATE_LINE" >> $TMPDIR/${IFACE_REPORT_PREFIX}-${IFACE}

			debug "check_service_settings() :: Missing line -> $TEMPLATE_LINE"
		fi
	done <  $DEV_TEMPLATE_DIR/services/$2
}

#
# Check the service configuration settings on interfaces
# Parameter 1: Absolute path to parsed interface file
# Parameter 2: Template file (absolute path)
#
check_service_exclude_settings()
{
	debug "check_service_exclude_settings()"

	# check if service template and definition file exists
	if [ ! -e "$DEV_TEMPLATE_DIR/services/$2.exclude" ]; then
		debug "Exclude file $DEV_TEMPLATE_DIR/services/$2.exclude not found"
		return
	fi

	while read TEMPLATE_LINE
	do
		# Skip comparation when "!" and empty line was found
		if [ -z "$TEMPLATE_LINE" ] || [ ${TEMPLATE_LINE:0:1} = '!' ]; then
			continue
		fi

		test -z "$TEMPLATE_LINE" && continue

		RESULT=$(grep "$TEMPLATE_LINE" $TMPDIR/$1)
		if [ -n "$RESULT" ]; then
			IFACE_E=$((IFACE_E+$(echo "$RESULT" | wc -l)))
			echo "$RESULT" | awk '{print "EXCLUDE: " $0}' >> $TMPDIR/${IFACE_REPORT_PREFIX}-${IFACE}

			debug "check_service_exclude_settings() :: Exclude line -> $RESULT"
		fi
	done <  $DEV_TEMPLATE_DIR/services/$2.exclude
}

#
# Check all ethernet and serial interfaces/service
#
function check_service_id() {

	debug "check_service_id()"
	
	# SEARCH=$(find $TMPDIR -iname "interface*ethernet*" -printf "%f\n" -or -iname "interface*serial*" -printf "%f\n" -or -iname "interface*vlan*" -printf "%f\n")
	SEARCH_IFACES=$(find $TMPDIR -iname "interface*" -printf "%f\n")

	# New select feature in 0.2.2v
	test $CHECK_IFACE_ETHERNET -eq 1 && SERVICE_ID_IFACE+="ethernet|"
	test $CHECK_IFACE_SERIAL -eq 1 && SERVICE_ID_IFACE+="_serial|"
	test $CHECK_IFACE_VLAN -eq 1 && SERVICE_ID_IFACE+="_vlan|"
	test $CHECK_IFACE_PORT_CHANNEL -eq 1 && SERVICE_ID_IFACE+="_Port-channel|"
	test $CHECK_IFACE_LOOPBACK -eq 1 && SERVICE_ID_IFACE+="_loopback|"
	test $CHECK_IFACE_TUNNEL -eq 1 && SERVICE_ID_IFACE+="_tunnel|"
	test $CHECK_IFACE_VIRT_TEMPLATE -eq 1 && SERVICE_ID_IFACE+="_Virtual-Template|"
	# Remove trailing '|' character

	SERVICE_ID_IFACES=$(echo $SERVICE_ID_IFACE | sed -e 's/|$//')
	
	SEARCH=$(echo "$SEARCH_IFACES" | egrep -i "$SERVICE_ID_IFACES" )

	for IFACE in $SEARCH; do
		# Detect interface type
		parse_iface $TMPDIR/$IFACE

		# Check if interface has NO description and is NOT shutdowned
		if [ "$IFACE_STATUS" = "UP" ] && [ -z "$IFACE_DESC" ]; then
			IFACE_NUM=$(($IFACE_NUM+1))
			IFACE_E=$((IFACE_E+1))
			echo "MISSING: description ... [SERVICE_ID]" >> $TMPDIR/${IFACE_REPORT_PREFIX}-${IFACE}
			continue
		fi

	
		if [ $IFACE_STATUS = "UP" ]; then
			# interface is UP

			IFACE_NUM=$(($IFACE_NUM+1))
	
			if [ $IFACE_TYPE != "N/A" ]; then
				# Interface type defined
				check_service_settings $IFACE $IFACE_TYPE 
				check_service_exclude_settings $IFACE $IFACE_TYPE 
			else
				# Interface type not defined
				IFACE_E=$((IFACE_E+1))

				# Fix interface description from filename "interface_TenGigabitEthernet9_0_0" -> "interface TenGigabitEthernet9/0/0"
				INTERFACE=$(echo "$IFACE" | sed -e 's/interface_/interface /' | tr '_' '/')

				echo "MISSING: description $IFACE_DESC ... [Service ID]" >> $TMPDIR/${IFACE_REPORT_PREFIX}-${IFACE}
			fi
		else
			# shutdowned interaface notification
			debug "$IFACE is shutdowned, skipping..."
		fi
		
	done
}

#
# Check ACL
#
function check_acl() {
	debug "check_acl()"

	# Check if "sections" directory exists
	if [ ! -e "$DEV_TEMPLATE_DIR/sections" ]; then
		echo "Section subdirectory does not exists, checking skipped"
		return
	fi

	# List all sessions
	SECTION_ACL_LIST=$(find "$DEV_TEMPLATE_DIR/sections/" \
		-type f -name "ip_access-list_*" -printf "%f\n" -o \
		-type l -name "ip_access-list_*" -printf "%f\n" -o \
		-type f -name "access-list_*" -printf "%f\n" -o \
		-type l -name "access-list_*" -printf "%f\n" -o \
		-type f -name "ipv6_access-list_*" -printf "%f\n" -o \
		-type l -name "ipv6_access-list_*" -printf "%f\n" -o \
		-type f -name "mac_access-list_*" -printf "%f\n" -o \
		-type l -name "mac_access-list_*" -printf "%f\n" | grep -v '.exclude')

	for SECTION_ACL in $SECTION_ACL_LIST; do
		debug "check_acl() :: ACL file $SECTION_ACL"

		SECTION_NUM=$(($SECTION_NUM+1))

		if [ ! -e $TMPDIR/$SECTION_ACL ]; then
			SECTION_ACL_E=$(($SECTION_ACL_E+1))
#			echo "warning: ACL \"$SECTION_ACL\" missing in configration file, skipping ..."
			echo "ERROR: Undefined ACL" > $TMPDIR/${ACL_REPORT_PREFIX}-$SECTION_ACL
			continue
		fi

		debug "check_acl() :: SECTION_ACL=$SECTION_ACL"
			
		RESULT=$(diff $TMPDIR/$SECTION_ACL $DEV_TEMPLATE_DIR/sections/$SECTION_ACL)

		if [ ! -z "$RESULT" ]; then
			echo "$RESULT" >> $TMPDIR/${ACL_REPORT_PREFIX}-$SECTION_ACL
			SECTION_ACL_E=$(($SECTION_ACL_E+1))
		fi

	done
}



#
# Print report of global configuration
#
function print_global_report() {

	# Print global configuration report
	if [ -e $GLOBAL_REPORT ]; then
		echo -e "\nGLOBAL configuration:\n=====================\n"
		if [ $COPY_AND_PASTE_MODE -eq 1 ]; then
			echo "---%<---"
			# Print pre-exec command(s)
			for ENTRY in "${COPY_AND_PASTE_PRE_EXEC[@]}"; do echo $ENTRY; done
			MISSING=$(egrep "^MISSING: " $GLOBAL_REPORT)
			if [ -n "$MISSING" ]; then
				echo "$MISSING" | sed -e 's/MISSING: / /' | tr -d '^$'
			fi
	
			EXCLUDE=$(egrep "^EXCLUDE: " $GLOBAL_REPORT)
			if [ -n "$EXCLUDE" ]; then
				echo "$EXCLUDE" | sed -e 's/EXCLUDE: / no /' | tr -d '^$'
			fi
			# Print post-exec command(s)
			for ENTRY in "${COPY_AND_PASTE_POST_EXEC[@]}"; do echo $ENTRY; done
			echo -e "---%<---"
		else
			MISSING=$(egrep "^MISSING: " $GLOBAL_REPORT)
			if [ -n "$MISSING" ]; then
				echo "($(echo "$MISSING" | wc -l)) Missing command(s):"
				echo "$MISSING" | sed -e 's/MISSING: /     /' | tr -d '^$'
			fi
			
			EXCLUDE=$(egrep "^EXCLUDE: " $GLOBAL_REPORT)
			if [ -n "$EXCLUDE" ]; then
				echo "($(echo "$EXCLUDE" | wc -l)) Exclude command(s):"
				echo "$EXCLUDE" | sed -e 's/EXCLUDE: /     /' | tr -d '^$'
			fi
		fi
	fi
}

#
# Print report of sections
#
function print_section_report() {
	local SEPARATOR=0

	if [ $SECTION_E -gt 0 ]; then
		echo -e "\nSECTION configuration:\n======================\n"
		if [ $COPY_AND_PASTE_MODE -eq 1 ]; then
			echo "---%<---"
			for REPORT in $(find $TMPDIR -name "${SECTION_REPORT_PREFIX}-*" -printf "%f\n"); do
				test $SEPARATOR -eq 1 && echo "!" || SEPARATOR=1
				# Print pre-exec command(s)
				for ENTRY in "${COPY_AND_PASTE_PRE_EXEC[@]}"; do echo $ENTRY; done

				echo "$(echo $REPORT | sed -e 's/report-section-/ /' | tr '_' ' ')"
		
				MISSING=$(egrep "^MISSING: " $TMPDIR/$REPORT)
				if [ -n "$MISSING" ]; then
					echo "$MISSING" | sed -e 's/MISSING: /  /' | tr -d '^$'
				fi
			
				EXCLUDE=$(egrep "^EXCLUDE: " $TMPDIR/$REPORT)
				if [ -n "$EXCLUDE" ]; then
					echo "$EXCLUDE" | sed -e 's/EXCLUDE: /  no /' | tr -d '^$'
				fi
		
				# Print post-exec command(s)
				for ENTRY in "${COPY_AND_PASTE_POST_EXEC[@]}"; do echo $ENTRY; done
			done
			echo "---%<---"
		else
	
		for REPORT in $(find $TMPDIR -name "${SECTION_REPORT_PREFIX}-*" -printf "%f\n"); do
			echo "Section: $(echo $REPORT | sed -e 's/report-section-//' | tr '_' ' ')"
	
			MISSING=$(egrep "^MISSING: " $TMPDIR/$REPORT)
			if [ -n "$MISSING" ]; then
				echo "($(echo "$MISSING" | wc -l)) Missing command(s):"
				echo "$MISSING" | sed -e 's/MISSING: /     /' | tr -d '^$'
			fi
		
			EXCLUDE=$(egrep "^EXCLUDE: " $TMPDIR/$REPORT)
			if [ -n "$EXCLUDE" ]; then
				echo "($(echo "$EXCLUDE" | wc -l)) Exclude command(s):"
				echo "$EXCLUDE" | sed -e 's/EXCLUDE: /     /' | tr -d '^$'
			fi
	
			echo
		done
		fi
	fi
}

#
# Print interface report
#
function print_service_id_report() {
	local SEPARATOR=0

	# Print service/interface report(s)
	if [ $IFACE_E -gt 0 ]; then
		echo -e "\nSERVICE configuration:\n======================\n"
		if [ $COPY_AND_PASTE_MODE -eq 1 ]; then
			echo "---%<---"
			for REPORT in $(find $TMPDIR -name "${IFACE_REPORT_PREFIX}-*" -printf "%f\n"); do
				test $SEPARATOR -eq 1 && echo "!" || SEPARATOR=1
				# Print pre-exec command(s)
				for ENTRY in "${COPY_AND_PASTE_PRE_EXEC[@]}"; do echo $ENTRY; done

				echo " interface $(echo $REPORT | sed -e 's/report-iface-interface_//' | tr '_' '/')"
		
				MISSING=$(egrep "^MISSING: " $TMPDIR/$REPORT)
				if [ -n "$MISSING" ]; then
					echo "$MISSING" | sed -e 's/MISSING: /  /' | tr -d '^$'
				fi
			
				EXCLUDE=$(egrep "^EXCLUDE: " $TMPDIR/$REPORT)
				if [ -n "$EXCLUDE" ]; then
					echo "$EXCLUDE" | sed -e 's/EXCLUDE: /  no /' | tr -d '^$'
				fi
		
				OTHER=$(egrep "^OTHER: " $TMPDIR/$REPORT)
				if [ -n "$OTHER" ]; then
					echo "$OTHER" | sed -e 's/OTHER: /  ! /' | tr -d '^$'
				fi

				# Print post-exec command(s)
				for ENTRY in "${COPY_AND_PASTE_POST_EXEC[@]}"; do echo $ENTRY; done
			done
			echo "---%<---"
		else
			for REPORT in $(find $TMPDIR -name "${IFACE_REPORT_PREFIX}-*" -printf "%f\n"); do
				echo "Interface $(echo $REPORT | sed -e 's/report-iface-interface_//' | tr '_' '/')"
		
				MISSING=$(egrep "^MISSING: " $TMPDIR/$REPORT)
				if [ -n "$MISSING" ]; then
					echo "($(echo "$MISSING" | wc -l)) Missing command(s):"
					echo "$MISSING" | sed -e 's/MISSING: /     /' | tr -d '^$'
				fi
			
				EXCLUDE=$(egrep "^EXCLUDE: " $TMPDIR/$REPORT)
				if [ -n "$EXCLUDE" ]; then
					echo "($(echo "$EXCLUDE" | wc -l)) Exclude command(s):"
						echo "$EXCLUDE" | sed -e 's/EXCLUDE: /     /' | tr -d '^$'
				fi
		
				OTHER=$(egrep "^OTHER: " $TMPDIR/$REPORT)
				if [ -n "$OTHER" ]; then
					echo "(1) Other note(s):"
					echo "$OTHER" | sed -e 's/OTHER: /     /' | tr -d '^$'
				fi
				echo
			done
		fi
	fi
}


#
# Print ACL report
#
function print_acl_report() {
	local SEPARATOR=0

	if [ $SECTION_ACL_E -gt 0 ]; then
		echo -e "\nACL configuration:\n==================\n"

		if [ $COPY_AND_PASTE_MODE -eq 1 ]; then
			echo "---%<---"

			for REPORT in $(find $TMPDIR -name "${ACL_REPORT_PREFIX}-*" -printf "%f\n"); do
				test $SEPARATOR -eq 1 && echo "!" || SEPARATOR=1
				# Print pre-exec command(s)
				for ENTRY in "${COPY_AND_PASTE_PRE_EXEC[@]}"; do echo $ENTRY; done
				
				RESULT=$(grep -c "ERROR: Undefined ACL" $TMPDIR/$REPORT) 
				if [ $RESULT -eq 0 ]; then
					echo " no $(echo $REPORT | sed -e 's/report-acl-\(.*\)/\1/' | sed -e 's/ip_access-list_standard_/ip access-list standard /' | sed -e 's/ip_access-list_extended_/ip access-list extended /' | sed -e 's/access-list_/access-list /' | sed -e 's/mac_access-list_/mac access-list /')"
				fi
				# Print access-list name if "ip_access-list"
				if [ $(echo "$REPORT" | grep -c 'ip_access-list') -eq 1 ]; then
					echo " $(echo $REPORT | sed -e 's/report-acl-\(.*\)/\1/' | sed -e 's/ip_access-list_standard_/ip access-list standard /' | sed -e 's/ip_access-list_extended_/ip access-list extended /' | sed -e 's/access-list_/access-list /' | sed -e 's/mac_access-list_/mac access-list /')"
				fi
				# Print content of the ACL
				cat $DEV_TEMPLATE_DIR/sections/$(echo $REPORT | sed -e 's/report-acl-//') | egrep -v "^$" | awk '{print "  " $0}'

				# Print post-exec command(s)
				for ENTRY in "${COPY_AND_PASTE_POST_EXEC[@]}"; do echo $ENTRY; done
			done
			echo "---%<---"
		else
			for REPORT in $(find $TMPDIR -name "${ACL_REPORT_PREFIX}-*" -printf "%f\n"); do
				RESULT=$(grep -c "ERROR: Undefined ACL" $TMPDIR/$REPORT) 
					echo "Section: $(echo $REPORT | sed -e 's/report-acl-\(.*\)/\1/' | sed -e 's/ip_access-list_standard_/ip access-list standard /' | sed -e 's/ip_access-list_extended_/ip access-list extended /' | sed -e 's/access-list_/access-list /' | sed -e 's/mac_access-list_/mac access-list /')"
				if [ $RESULT -eq 1 ]; then
					echo "(1) ACL not defined. Original ACL template:"
				else
					echo "(1) Original ACL template:"
				fi
				cat $DEV_TEMPLATE_DIR/sections/$(echo $REPORT | sed -e 's/report-acl-//') | egrep -v "^$" | awk '{print "     " $0}'
				echo
			done
		fi
	fi
}

#
# Read CTM configuration file
#
function read_config_file() {

	# Read configuration file
	if [ ! -s "$CONFIGFILE" ]; then 
		echo "ERROR: CTM configuration file '$CONFIGFILE' not found or is empty"
		exit 1
	fi
	
	# Read configuration file
	. $CONFIGFILE

	if [ -z "$TEMPLATE_DIR" ] || [ -z "$CHECK_SECTIONS" ]; then
		echo "ERROR: Configuration file '$CONFIGFILE' is incomplete. Missing some arguments."
		exit 1
	fi
}

#
# Check CTM input arguments
#
function check_arguments() {

	# Check Cisco configuration file if exist and has non-zero size
	if [ ! -s "$CTM_CISCO_CONFIG" ]; then 
		echo "ERROR: Cisco configuration file '$CTM_CISCO_CONFIG' not found or is empty"
		exit 1
	fi
	
	# Check if device type directory exists
	if [ ! -e "$TEMPLATE_DIR/$CTM_DEVICE" ]; then 
		echo "ERROR: Device type '$CTM_DEVICE' does not exists. Check template(s) directory $TEMPLATE_DIR/..."
		exit 1
	fi

	# Check if role directory exists
	if [ ! -e "$TEMPLATE_DIR/$CTM_DEVICE/$CTM_ROLE" ]; then 
		echo "ERROR: Device type '$CTM_ROLE' does not exists. Check template(s) directory $TEMPLATE_DIR/$CTM_DEVICE/..."
		exit 1
	fi
}

#
# Cleanup temporary files
#
function cleanup() {
	debug "Cleaning up temporary files"
	if [ "$(echo "$1" | egrep -ic '^nocleanup$')" -eq 0 ]; then
		test -e $TMPDIR && rm -f $TMPDIR/*
		rmdir $TMPDIR
	fi
}



##############################################################################
#                              MAIN CTM Script                               #
##############################################################################

test -n "$DEBUG" && echo "NOTE: DEBUG mode enabled, loggin to $DEBUGFILE file (lower performance)"

read_config_file

if [ ! $# -ge 3 ]; then
	show_help
	exit
fi

# Detect mode
if [ "$(echo "$4" | egrep -ic '^parseonly$')" -eq 1 ]; then
	MODE="PARSEONLY"
fi 

CTM_DEVICE="$1"
CTM_ROLE="$2"
CTM_CISCO_CONFIG="$3"

check_arguments

# Create temporary directories before startup
test ! -e $TMPDIR && mkdir $TMPDIR 

remove_trailing_characters $CTM_CISCO_CONFIG

parse_config_file $1

if [ "$MODE" = "PARSEONLY" ]; then
	echo "Configuration file '$CTM_CISCO_CONFIG' is parsed to $TMPDIR directory" 
	exit
fi

	DEV_TEMPLATE_DIR="$TEMPLATE_DIR/$CTM_DEVICE/$CTM_ROLE"

	if [ ! -e "$DEV_TEMPLATE_DIR" ]; then
		echo "Error: Template directory $DEV_TEMPLATE_DIR does not exist"
		cleanup $4
		exit 1
	fi

echo "Hostname: $DEV_HOSTNAME, Device type: ${CTM_DEVICE}, Role: ${CTM_ROLE}"
# Check global settings
test $CHECK_GLOBAL -eq 1 && check_global_settings 
# Check for excluded commands
test $CHECK_GLOBAL_EXCLUDE -eq 1 && check_global_exclude_settings 
# Check sections
test $CHECK_SECTIONS -eq 1 && check_sections
# Check "excluded" commands in sections
test $CHECK_SECTIONS_EXCLUDE -eq 1 && check_sections_exclude
# Check interface configuration 
test $CHECK_IFACE_SERVICE_ID -eq 1 && check_service_id

check_acl

#
# Print reports
#

echo -e "\nGlobal error(s): $GLOBAL_E, Section(s): ${SECTION_NUM} err: $((${SECTION_E}+${SECTION_ACL_E})), Iface(s): ${IFACE_NUM} err: ${IFACE_E}\n"

print_global_report

print_section_report

print_acl_report

print_service_id_report
	

# Remove unused files
cleanup $4
